; TWI non-interrupt master-mode for Dallas DS1307 RTC, Mega128
; TWI registers are in memory space (LDS, STS) on this CPU
;
;	.NOLIST
;	.INCLUDE "m128def.inc"
;	.LIST

; Initial code by Peret
; Modified by John Samperi Ampertronics P/L 30/6/2013
; Added wait_for_TX_stop and wait_for_RX_stop as the code would lock at times.
; Added automatic TWI_TWBR calculations with the following
; 
#define TWI_scl_bitrate_l 100000				; Desired TWI low speed bit rate ****
#define TWI_TWPS 0x00							; TWI prescaler
#define TWI_TWBR_L (fosc/TWI_scl_bitrate_l-16)/2 ; Works out TWBR low speed value


	.LISTMAC

	.macro	ini
	.if	@1<0x40
	in	@0,@1
	.else
	lds	@0,@1
	.endif
	.endm

	.macro	outi
	.if	@0<0x40
	out	@0,@1
	.else
	sts	@0,@1
	.endif
	.endm

.def	acc=r16
.def	acchi=r17
.def	err_flags=r18
.def	count=r19
.equ	device_id	=(0x68<<1)	; 68 for DS1307 real time clock
					; - must match 7-bit device ID

; Initialize TWI - run me first
twi_init:
	ldi		acc,TWI_TWBR_L		; Set bit rate register for low speed
	outi	TWBR,acc
	clr		acc
	outi	TWSR,acc			; no prescaler
	ldi		acc,(1<<TWINT | 1<<TWSTO | 1<<TWEN)
	outi	TWCR,acc			; enable and send stop
	cbr		err_flags,0x01		; clear TWI error flag
	ret

; Leaves TWI enabled, SCL low, TWINT set

; General TWI write
; Write (count) bytes from address (Y) starting at TWI address (acc)
twi_write:
	push	acchi
	push	acc
	cbr		err_flags,0x01	; clear TWI error flag
	rcall	twi_start		; send start condition
	cpi		acc,0x08		; must be 8
	brne	twiw5			; else error

; send device address and write flag
	ldi		acc,device_id	; ID + write
	rcall	twi_wdat		; send data byte
	cpi		acc,0x18		; must be 18
	brne	twiw5			; else error

; send one-byte address
; larger devices such as EEPROMs may need two or more address
; bytes to be sent here
	pop		acc				; recover TWI address
	push	acc
	andi	acc,0x3F
	rcall	twi_wdat
	cpi		acc,0x28		; must be 28
	brne	twiw5			; else error

; send (count) data bytes
twiw4:	
	ld		acc,Y+
	rcall	twi_wdat
	cpi		acc,0x28		; must be 28
	brne	twiw5			; else error
	dec		count
	brne	twiw4			; do until done
	rjmp	twiw6

twiw5:	
	sbr	err_flags,0x01		; set error flag

; send stop condition
twiw6:	
	ldi		acc,(1<<TWINT | 1<<TWSTO | 1<<TWEN)
	outi	TWCR,acc		; send stop condition

;JS
wait_for_TX_stop:
	ini		acc,TWCR
	sbrc	acc,TWSTO
	rjmp	wait_for_TX_stop ; wait for stop bit to clear

	pop		acc
	pop		acchi
	ret


; General TWI read
; Read (count) bytes from TWI address (acc) to address (Y)
twi_read:
	push	acchi
	push	acc
	cbr		err_flags,0x01	; clear TWI error flag

	rcall	twi_start
	cpi		acc,0x08		; must be 8
	brne	twiw5			; else error

; send device address and write flag
	ldi		acc,device_id	; ID + write
	rcall	twi_wdat		; send data byte
	cpi		acc,0x18		; must be 18
	brne	twiw5			; else error

; send byte address
; again, more than one byte may be needed for memories
	pop	acc
	push	acc
	andi	acc,0x3F
	rcall	twi_wdat
	cpi		acc,0x28		; must be 28
	brne	twiw5			; else error

; Send repeated start to terminate write
	rcall	twi_start
	cpi		acc,0x10		; must be 10
	brne	twiw5			; else error

; send device address and read flag
	ldi		acc,device_id+1	; ID + read
	rcall	twi_wdat		; send data byte
	cpi		acc,0x40		; must be 40
	brne	twiw5			; else error

; receive a byte
	ldi		acchi,0x01		; for CPSE
twir1:	
	ldi		acc,(1<<TWINT | 1<<TWEN)
	cpse	count,acchi		; on all except last byte ..
	sbr		acc,(1<<TWEA)	; also set ack enable
	outi	TWCR,acc		; read command
twir2:	
	ini		acc,TWCR
	sbrs	acc,TWINT
	rjmp	twir2			; wait for completion

	ini		acc,TWSR		; check the status
	andi	acc,0xF8
	cpi		acc,0x58		; note: will be 58 on last byte
	breq	twir3
	cpi		acc,0x50		; must be 50
	brne	twiw5			; else error
twir3:
	ini		acc,TWDR		; get the byte
	st		Y+,acc
	dec		count
	brne	twir1

; send stop condition
	ldi		acc,(1<<TWINT | 1<<TWSTO | 1<<TWEN)
	outi	TWCR,acc		; send stop condition

;JS
wait_for_RX_stop:
	ini		acc,TWCR
	sbrc	acc,TWSTO
	rjmp	wait_for_RX_stop ; wait for stop bit to clear

	pop		acc
	pop		acchi
	ret

; send start condition
twi_start:
; clear TWINT, set TWSTA and TWEN
	ldi		acc,(1<<TWINT | 1<<TWSTA | 1<<TWEN)
	outi	TWCR,acc

twis2:	
	ini		acc,TWCR
	sbrs	acc,TWINT
	rjmp	twis2			; wait for completion
	ini		acc,TWSR		; check the status
	andi	acc,0xF8
	ret

; send a data byte
twi_wdat:
	ini		acchi,TWCR		; wait for TWINT high
	sbrs	acchi,TWINT		; or it will lock up
	rjmp	twi_wdat

	outi	TWDR,acc
	ldi		acc,(1<<TWINT | 1<<TWEN)
	outi	TWCR,acc
twiw2:	
	ini		acc,TWCR
	sbrs	acc,TWINT
	rjmp	twiw2			; wait for completion

	ini		acc,TWSR		; check the status
	andi	acc,0xF8
	ret

